@@ -3,11 +3,11 @@
 
 package Net::SNMP::Security::USM;
 
-# $Id: USM.pm,v 2.2 2003/09/11 19:14:42 dtown Exp $
+# $Id: USM.pm,v 2.4 2005/07/20 13:53:07 dtown Exp $
 
 # Object that implements the SNMPv3 User-based Security Model.
 
-# Copyright (c) 2001-2003 David M. Town <dtown@cpan.org>
+# Copyright (c) 2001-2005 David M. Town <dtown@cpan.org>
 # All rights reserved.
 
 # This program is free software; you may redistribute it and/or modify it
@@ -17,14 +17,11 @@ package Net::SNMP::Security::USM;
 
 use strict;
 
-require Net::SNMP::Security;
+use Net::SNMP::Security qw( :ALL );
 
 use Net::SNMP::Message qw(
-   :msgFlags :securityLevels asn1_itoa
-   OCTET_STRING SEQUENCE INTEGER SNMP_VERSION_3 SECURITY_MODEL_USM TRUE FALSE
-); 
-
-use Sys::Hostname qw(hostname);
+   :msgFlags asn1_itoa OCTET_STRING SEQUENCE INTEGER SNMP_VERSION_3 TRUE FALSE
+);
 
 use Crypt::DES();
 use Digest::MD5();
@@ -33,55 +30,55 @@ use Digest::HMAC();
 
 ## Version of the Net::SNMP::Security::USM module
 
-our $VERSION = v2.0.2;
-
-## Package variables
-
-our $ENGINE_ID;      # Our authoritativeEngineID 
-
-our $DEBUG = FALSE;  # Debug flag
+our $VERSION = v3.0.0;
 
 ## Handle importing/exporting of symbols
 
 use Exporter();
 
-our @ISA = qw(Exporter Net::SNMP::Security);
+our @ISA = qw( Net::SNMP::Security Exporter );
 
-our @EXPORT_OK = qw(
-   AUTH_PROTOCOL_NONE AUTH_PROTOCOL_HMACMD5 AUTH_PROTOCOL_HMACSHA
-   PRIV_PROTOCOL_NONE PRIV_PROTOCOL_DES PRIV_PROTOCOL_3DESEDE
-   PRIV_PROTOCOL_AESCFB128 PRIV_PROTOCOL_AESCFB192 PRIV_PROTOCOL_AESCFB256
-   SECURITY_LEVEL_NOAUTHNOPRIV SECURITY_LEVEL_AUTHNOPRIV 
-   SECURITY_LEVEL_AUTHPRIV
-); 
+our @EXPORT_OK;
 
 our %EXPORT_TAGS = (
-   levels      => [
+   authprotos => [
+      qw( AUTH_PROTOCOL_NONE AUTH_PROTOCOL_HMACMD5 AUTH_PROTOCOL_HMACSHA )
+   ],
+   levels     => [
       qw( SECURITY_LEVEL_NOAUTHNOPRIV SECURITY_LEVEL_AUTHNOPRIV
           SECURITY_LEVEL_AUTHPRIV )
    ],
-   authprotos  => [
-      qw( AUTH_PROTOCOL_NONE AUTH_PROTOCOL_HMACMD5 AUTH_PROTOCOL_HMACSHA )
+   models     => [
+      qw( SECURITY_MODEL_ANY SECURITY_MODEL_SNMPV1 SECURITY_MODEL_SNMPV2C
+          SECURITY_MODEL_USM )
    ],
-   privprotos  => [
-      qw( PRIV_PROTOCOL_NONE PRIV_PROTOCOL_DES PRIV_PROTOCOL_3DESEDE 
-          PRIV_PROTOCOL_AESCFB128 PRIV_PROTOCOL_AESCFB192 
-          PRIV_PROTOCOL_AESCFB256 )
-   ],
-   ALL         => [@EXPORT_OK]
+   privprotos => [
+      qw( PRIV_PROTOCOL_NONE PRIV_PROTOCOL_DES PRIV_PROTOCOL_AESCFB128
+          PRIV_PROTOCOL_DRAFT_3DESEDE PRIV_PROTOCOL_DRAFT_AESCFB128
+          PRIV_PROTOCOL_DRAFT_AESCFB192 PRIV_PROTOCOL_DRAFT_AESCFB256 )  
+   ]
 );
 
-## Authentication protocols
+Exporter::export_ok_tags( qw( authprotos levels models privprotos ) );
+
+$EXPORT_TAGS{ALL} = [ @EXPORT_OK ];
+
+## RCC 3414 - Authentication protocols
 
 sub AUTH_PROTOCOL_NONE()    { '1.3.6.1.6.3.10.1.1.1' } # usmNoAuthProtocol
 sub AUTH_PROTOCOL_HMACMD5() { '1.3.6.1.6.3.10.1.1.2' } # usmHMACMD5AuthProtocol
 sub AUTH_PROTOCOL_HMACSHA() { '1.3.6.1.6.3.10.1.1.3' } # usmHMACSHAAuthProtocol
 
-## Privacy protocols
+## RFC 3414 - Privacy protocols
 
 sub PRIV_PROTOCOL_NONE()    { '1.3.6.1.6.3.10.1.2.1' } # usmNoPrivProtocol
 sub PRIV_PROTOCOL_DES()     { '1.3.6.1.6.3.10.1.2.2' } # usmDESPrivProtocol
 
+## RFC 3826 - The AES Cipher Algorithm in the SNMP USM 
+
+# usmAesCfb128Protocol
+sub PRIV_PROTOCOL_AESCFB128()        {  '1.3.6.1.6.3.10.1.2.4' }
+
 # The privacy protocols below have been implemented using the draft 
 # specifications intended to extend the User-based Security Model 
 # defined in RFC 3414.  Since the object definitions have not been 
@@ -92,56 +89,23 @@ sub PRIV_PROTOCOL_DES()     { '1.3.6.1.6.3.10.1.2.2' } # usmDESPrivProtocol
 # Reeder and Gudmunsson; October 1999, expired April 2000 
 
 # usm3DESPrivProtocol 
-sub PRIV_PROTOCOL_3DESEDE()   { '1.3.6.1.4.1.14832.1.1' }
+sub PRIV_PROTOCOL_DRAFT_3DESEDE()    { '1.3.6.1.4.1.14832.1.1' }
 
 # AES Cipher Algorithm in the USM <draft-blumenthal-aes-usm-04.txt>
 # Blumenthal, Maino, and McCloghrie; October 2002, expired April 2003 
 
-# usmAESCfb128PrivProtocol
-sub PRIV_PROTOCOL_AESCFB128() { '1.3.6.1.4.1.14832.1.2' } 
+# usmAESCfb128PrivProtocol 
+sub PRIV_PROTOCOL_DRAFT_AESCFB128()  { '1.3.6.1.4.1.14832.1.2' } 
 
-# usmAESCfb192PrivProtocol
-sub PRIV_PROTOCOL_AESCFB192() { '1.3.6.1.4.1.14832.1.3' }
+# usmAESCfb192PrivProtocol 
+sub PRIV_PROTOCOL_DRAFT_AESCFB192()  { '1.3.6.1.4.1.14832.1.3' } 
 
 # usmAESCfb256PrivProtocol
-sub PRIV_PROTOCOL_AESCFB256() { '1.3.6.1.4.1.14832.1.4' } 
-                                                         
-our $NO_AES_SUPPORT;
+sub PRIV_PROTOCOL_DRAFT_AESCFB256()  { '1.3.6.1.4.1.14832.1.4' } 
 
-BEGIN
-{
-   # Attempt to load the Crypt::Rijndael module.  If this module is 
-   # not found, disable support for the AES Cipher Algorithm.
-
-   if (!eval('require Crypt::Rijndael')) {
-      if ($@ =~ /(\S+\.pm)/) {
-         $NO_AES_SUPPORT = sprintf(' (Required module %s not found)', $1);
-      } else {
-         $NO_AES_SUPPORT = sprintf(' (%s)', $@);
-      }
-      # Define the MODE_CFB constant to prevent a "strict subs" error.
-      *Crypt::Rijndael::MODE_CFB = sub { 3 }; 
-   }
-
-   # Initialize our snmpEngineID using the algorithm described
-   # in RFC 3411.
-
-   # SnmpEngineID ::= TEXTUAL-CONVENTION
-
-   # The first bit is set to one to indicate that the RFC 3411 
-   # algorithm is being used.  The first fours bytes are to be
-   # the agent's SNMP management private enterprise number, but
-   # they are set to all zeros. The fifth byte is set to one to
-   # indicate that the final four bytes are a IPv4 address.
-
-   eval { 
-   $ENGINE_ID = pack('H10', '8000000001') . scalar(gethostbyname(hostname()));
-   };
-
-   # Fallback in case gethostbyname() or hostname() fail
-   $ENGINE_ID = pack('x11H2', '01') if ($@);
-}
+## Package variables
 
+our $ENGINE_ID;               # Our authoritative snmpEngineID                                                         
 # [public methods] -----------------------------------------------------------
 
 sub new
@@ -217,14 +181,10 @@ sub new
 
    }
 
-   # Set the authoritativeEngineID if the user did not provide one
-
-   if ($this->{_authoritative}) {
-      $this->{_engine_id}    = $ENGINE_ID if ($this->{_engine_id} eq '');
-      $this->{_engine_boots} = 1;
-      $this->{_synchronized} = TRUE;
-      $this->{_discovered}   = TRUE;
-   }
+   # Generate a snmpEngineID and populate the object accordingly
+   # if we are an authoritative snmpEngine.
+ 
+   $this->_snmp_engine_init if ($this->{_authoritative});
 
    # Define the securityParameters
    if (!defined($this->_security_params)) {
@@ -244,6 +204,18 @@ sub generate_request_msg
 
    return $this->_error('Required PDU and/or Message missing') unless (@_ == 3);
 
+   # Validate the SNMP version of the PDU
+   if ($pdu->version != $this->{_version}) {
+      return $this->_error('Invalid version [%d]', $pdu->version);
+   }
+
+   # Validate the securityLevel of the PDU
+   if ($pdu->security_level > $this->{_security_level}) {
+      return $this->_error(
+         'Unsupported securityLevel [%d]', $pdu->security_level
+      );
+   }
+
    # Validate PDU type with snmpEngine type
    if ($pdu->expect_response) {
       if ($this->{_authoritative}) {
@@ -279,7 +251,7 @@ sub generate_request_msg
    my $priv_params = '';
 
    # encryptedPDU::=OCTET STRING
-   if ($this->{_security_level} > SECURITY_LEVEL_AUTHNOPRIV) {
+   if ($pdu->security_level > SECURITY_LEVEL_AUTHNOPRIV) {
       if (!defined($this->_encrypt_data($msg, $priv_params, $pdu_buffer))) {
          return $this->_error;
       }
@@ -295,7 +267,7 @@ sub generate_request_msg
    my $auth_params = '';
    my $auth_location = 0;
 
-   if ($this->{_security_level} > SECURITY_LEVEL_NOAUTHNOPRIV) {
+   if ($pdu->security_level > SECURITY_LEVEL_NOAUTHNOPRIV) {
    
       # Save the location to fill in msgAuthenticationParameters later
       $auth_location = $msg->length + 12 + length($pdu_buffer);
@@ -309,7 +281,7 @@ sub generate_request_msg
    }
 
    # msgUserName::=OCTET STRING 
-   if (!defined($msg->prepare(OCTET_STRING, $this->_user_name))) {
+   if (!defined($msg->prepare(OCTET_STRING, $pdu->security_name))) {
       return $this->_error($msg->error);
    } 
 
@@ -329,7 +301,7 @@ sub generate_request_msg
    }
 
    # UsmSecurityParameters::= SEQUENCE
-   if (!defined($msg->prepare(SEQUENCE, $msg->clear))) {
+   if (!defined($msg->prepare(SEQUENCE))) {
       return $this->_error($msg->error);
    }
 
@@ -349,17 +321,17 @@ sub generate_request_msg
    }
 
    # version::=INTEGER
-   if (!defined($msg->prepare(INTEGER, $pdu->version))) {
+   if (!defined($msg->prepare(INTEGER, $this->{_version}))) {
       return $this->_error($msg->error);
    }
 
    # message::=SEQUENCE
-   if (!defined($msg->prepare(SEQUENCE, $msg->clear))) {
+   if (!defined($msg->prepare(SEQUENCE))) {
       return $this->_error($msg->error);
    }
 
    # Apply authentication
-   if ($this->{_security_level} > SECURITY_LEVEL_NOAUTHNOPRIV) {
+   if ($pdu->security_level > SECURITY_LEVEL_NOAUTHNOPRIV) {
       if (!defined($this->_authenticate_outgoing_msg($msg, $auth_location))) {
          return $this->_error($msg->error);
       }
@@ -422,8 +394,7 @@ sub process_incoming_msg
    }
 
    # msgUserName::=OCTET STRING (SIZE(0..32))
-   my $msg_user_name;
-   if (!defined($msg_user_name = $msg->process(OCTET_STRING))) {
+   if (!defined($msg->security_name($msg->process(OCTET_STRING)))) {
       return $this->_error($msg->error); 
    }
 
@@ -463,9 +434,9 @@ sub process_incoming_msg
          );
       }
 
-      if ($msg_user_name ne $this->_user_name) {
+      if ($msg->security_name ne $this->_user_name) {
          return $this->_error(
-            'Unknown incoming msgUserName [%s]', $msg_user_name
+            'Unknown incoming msgUserName [%s]', $msg->security_name 
          );
       }
 
@@ -478,38 +449,15 @@ sub process_incoming_msg
 
    }
 
-   # Get the securityLevel of the incoming message from the msgFlags
-
-   my $security_level = SECURITY_LEVEL_NOAUTHNOPRIV;
-   my $msg_flags = $msg->msg_flags;
-
-   if ($msg_flags & MSG_FLAGS_AUTH) {
-
-      $security_level = SECURITY_LEVEL_AUTHNOPRIV;
-
-      if ($msg_flags & MSG_FLAGS_PRIV) {
-         $security_level = SECURITY_LEVEL_AUTHPRIV;
-      }
-
-   } elsif ($msg_flags & MSG_FLAGS_PRIV) {
-
-      # RFC 3412 - Section 7.2 1d: "If the authFlag is not set  
-      # and privFlag is set... ...the message is discarded..."
-
-      return $this->_error('Invalid incoming msgFlags [0x%02x]', $msg_flags);
-   } 
-
-   # RFC 3412 - Section 7.2 1e: "Any other bits... ...are ignored."
-   if ($msg_flags & ~MSG_FLAGS_MASK) {
-      DEBUG_INFO('questionable msgFlags [0x%02x]', $msg_flags);
-   }
-
    # Validate the incoming securityLevel
 
-   if ($security_level > $this->{_security_level}) {
-      return $this->_error('Unsupported securityLevel [%d]', $security_level);
-   }
+   my $security_level = $msg->security_level;
 
+   if ($security_level > $this->{_security_level}) {
+      return $this->_error(
+          'Unsupported incoming securityLevel [%d]', $security_level
+      );
+   }
  
    if ($security_level > SECURITY_LEVEL_NOAUTHNOPRIV) {
 
@@ -546,9 +494,9 @@ sub process_incoming_msg
          # field..."  We must prepend the snmpEngineBoots and
          # snmpEngineTime as received in order to compute the IV.
 
-         if (($this->{_priv_protocol} eq PRIV_PROTOCOL_AESCFB128) ||
-             ($this->{_priv_protocol} eq PRIV_PROTOCOL_AESCFB192) ||
-             ($this->{_priv_protocol} eq PRIV_PROTOCOL_AESCFB256))
+         if (($this->{_priv_protocol} eq PRIV_PROTOCOL_AESCFB128)       ||
+             ($this->{_priv_protocol} eq PRIV_PROTOCOL_DRAFT_AESCFB192) ||
+             ($this->{_priv_protocol} eq PRIV_PROTOCOL_DRAFT_AESCFB256))
          {
             substr($priv_params, 0, 0) = pack(
                'NN', $msg_engine_boots, $msg_engine_time
@@ -579,8 +527,10 @@ sub user_name
 
 sub auth_protocol
 {
-   if ($_[0]->{_security_level} > SECURITY_LEVEL_NOAUTHNOPRIV) {
-      $_[0]->{_auth_protocol};
+   my ($this) = @_;
+
+   if ($this->{_security_level} > SECURITY_LEVEL_NOAUTHNOPRIV) {
+      $this->{_auth_protocol};
    } else {
       AUTH_PROTOCOL_NONE;
    }
@@ -593,8 +543,10 @@ sub auth_key
 
 sub priv_protocol
 {
-   if ($_[0]->{_security_level} > SECURITY_LEVEL_AUTHNOPRIV) {
-      $_[0]->{_priv_protocol};
+   my ($this) = @_;
+
+   if ($this->{_security_level} > SECURITY_LEVEL_AUTHNOPRIV) {
+      $this->{_priv_protocol};
    } else {
       PRIV_PROTOCOL_NONE;
    }
@@ -632,89 +584,136 @@ sub security_model
    SECURITY_MODEL_USM;
 }
 
-sub discovered
+sub security_name
 {
-   if ($_[0]->{_security_level} > SECURITY_LEVEL_NOAUTHNOPRIV) {
-      ($_[0]->{_discovered} && $_[0]->{_synchronized});
-   } else {
-      $_[0]->{_discovered};
-   }
+   $_[0]->_user_name;
 }
 
-sub debug
+sub discovered
 {
-   (@_ == 2) ? $DEBUG = ($_[1]) ? TRUE : FALSE : $DEBUG;
+   my ($this) = @_;
+
+   if ($this->{_security_level} > SECURITY_LEVEL_NOAUTHNOPRIV) {
+      ($this->{_discovered} && $this->{_synchronized});
+   } else {
+      $this->{_discovered};
+   }
 }
 
 # [private methods] ----------------------------------------------------------
 
 sub _version
 {
-   if ($_[1] != SNMP_VERSION_3) {
-      return $_[0]->_error('Invalid SNMP version specified [%s]', $_[1]);
+   my ($this, $version) = @_;
+
+   if ($version != SNMP_VERSION_3) {
+      return $this->_error('Invalid SNMP version specified [%s]', $version);
    }
 
-   $_[0]->{_version} = $_[1];
+   $this->{_version} = $version;
 }
 
 sub _engine_id
 {
+   my ($this, $engine_id) = @_;
+
    if (@_ == 2) {
-      if ($_[1] =~ /^(?i:0x)?([a-fA-F0-9]{10,64})$/) {
-         $_[0]->{_engine_id} = pack('H*', length($1) % 2 ? '0'.$1 : $1);
+      if ($engine_id =~ /^(?i:0x)?([a-fA-F0-9]{10,64})$/) {
+         $this->{_engine_id} = pack('H*', length($1) % 2 ? '0'.$1 : $1);
       } else {
-         return $_[0]->_error('Invalid authoritativeEngineID format specified');
+         return $this->_error('Invalid authoritativeEngineID format specified');
       }
    }
 
-   $_[0]->{_engine_id};
+   $this->{_engine_id};
 }
 
 sub _user_name
 {
+   my ($this, $user_name) = @_;
+
    if (@_ == 2) {
-      if ($_[1] eq '') {
-         return $_[0]->_error('Empty userName specified');
-      } elsif (length($_[1]) > 32) {
+      if ($user_name eq '') {
+         return $this->_error('Empty userName specified');
+      } elsif (length($user_name) > 32) {
          return $_[0]->_error(
             'Invalid userName length [%d octet%s]', 
-            length($_[1]), length($_[1]) != 1 ? 's' : ''
+            length($user_name), length($user_name) != 1 ? 's' : ''
          );
       }
-      $_[0]->{_user_name} = $_[1];
+      $this->{_user_name} = $user_name;
    }
 
    # RFC 3414 Section 4 - "Discovery... ...msgUserName of zero-length..."
 
-   ($_[0]->{_discovered}) ? $_[0]->{_user_name} : '';
+   ($this->{_discovered}) ? $this->{_user_name} : '';
+}
+
+sub _snmp_engine_init
+{
+   my ($this) = @_;
+
+   if ($this->{_engine_id} eq '') {
+
+      # Initialize our snmpEngineID using the algorithm described 
+      # in RFC 3411 - SnmpEngineID::=TEXTUAL-CONVENTION.
+
+      # The first bit is set to one to indicate that the RFC 3411
+      # algorithm is being used.  The first fours bytes are to be
+      # the agent's SNMP management private enterprise number, but
+      # they are set to all zeros. The fifth byte is set to one to
+      # indicate that the final four bytes are an IPv4 address.
+
+      if (!defined($ENGINE_ID)) {
+         eval {
+            require Sys::Hostname;
+            $ENGINE_ID = pack('H10', '8000000001') .
+                         scalar(gethostbyname(Sys::Hostname::hostname()));
+         };
+
+         # Fallback in case gethostbyname() or hostname() fail
+         $ENGINE_ID = pack('x11H2', '01') if ($@);    
+      }
+
+      $this->{_engine_id} = $ENGINE_ID;
+   }
+
+   $this->{_engine_boots} = 1;
+   $this->{_time_epoc}    = $^T;
+   $this->{_synchronized} = TRUE;
+   $this->{_discovered}   = TRUE;
 }
 
 sub _auth_key
 {
+   my ($this, $auth_key) = @_;
+
    if (@_ == 2) {
-      if ($_[1] =~ /^(?i:0x)?([a-fA-F0-9]+)$/) {
-         $_[0]->{_auth_key} = pack('H*', length($1) % 2 ? '0'.$1 : $1); 
-         if (!defined($_[0]->_auth_key_validate)) {
-            return $_[0]->_error;
+      if ($auth_key =~ /^(?i:0x)?([a-fA-F0-9]+)$/) {
+         $this->{_auth_key} = pack('H*', length($1) % 2 ? '0'.$1 : $1); 
+         if (!defined($this->_auth_key_validate)) {
+            return $this->_error;
          }
       } else {
-         return $_[0]->_error('Invalid authKey format specified');
+         return $this->_error('Invalid authKey format specified');
       }
    }
 
-   $_[0]->{_auth_key};
+   $this->{_auth_key};
 }
 
 sub _auth_password
 {
+   my ($this, $auth_password) = @_;
+
    if (@_ == 2) {
-      if ($_[1] eq '') {
+      if ($auth_password eq '') {
          return $_[0]->_error('Empty authentication password specified');
       } 
-      $_[0]->{_auth_password} = $_[1];
+      $this->{_auth_password} = $auth_password;
    }
 
-   $_[0]->{_auth_password};
+   $this->{_auth_password};
 }
 
 sub _auth_protocol
@@ -740,9 +739,7 @@ sub _auth_protocol
 
       if (@match > 1) {
          return $this->_error('Ambiguous authProtocol specified [%s]', $proto);
-      }
-
-      if (@match != 1) {
+      } elsif (@match != 1) {
          return $this->_error('Unknown authProtocol specified [%s]', $proto);
       }
 
@@ -754,44 +751,48 @@ sub _auth_protocol
 
 sub _priv_key
 {
+   my ($this, $priv_key) = @_;
+
    if (@_ == 2) {
 
-      if ($_[1] =~ /^(?i:0x)?([a-fA-F0-9]+)$/) {
+      if ($priv_key =~ /^(?i:0x)?([a-fA-F0-9]+)$/) {
 
-         $_[0]->{_priv_key} = pack('H*', length($1) % 2 ? '0'.$1 : $1);
+         $this->{_priv_key} = pack('H*', length($1) % 2 ? '0'.$1 : $1);
 
          # For backwards compatibility with previous versions of
          # this module truncate 20 byte SHA1 keys to 16 bytes. 
 
-         if (length($_[0]->{_priv_key}) == 20) {
-            $_[0]->{_priv_key} = substr($_[0]->{_priv_key}, 0, 16);
+         if (length($this->{_priv_key}) == 20) {
+            $this->{_priv_key} = substr($this->{_priv_key}, 0, 16);
          }
 
-         if (!defined($_[0]->_priv_key_validate)) {
-            return $_[0]->_error;
+         if (!defined($this->_priv_key_validate)) {
+            return $this->_error;
          }
 
       } else {
 
-         return $_[0]->_error('Invalid privKey format specified');
+         return $this->_error('Invalid privKey format specified');
 
       }
 
    }
 
-   $_[0]->{_priv_key};
+   $this->{_priv_key};
 }
 
 sub _priv_password
 {
+   my ($this, $priv_password) = @_;
+
    if (@_ == 2) {
-      if ($_[1] eq '') {
-         return $_[0]->_error('Empty privacy password specified');
+      if ($priv_password eq '') {
+         return $this->_error('Empty privacy password specified');
       }
-      $_[0]->{_priv_password} = $_[1];
+      $this->{_priv_password} = $priv_password;
    }
 
-   $_[0]->{_priv_password};
+   $this->{_priv_password};
 }
 
 sub _priv_protocol
@@ -801,21 +802,28 @@ sub _priv_protocol
    if (@_ == 2) {
 
       my $protocols = {
-         'des',                   PRIV_PROTOCOL_DES,
-         'cbcdes',                PRIV_PROTOCOL_DES,
-         PRIV_PROTOCOL_DES,       PRIV_PROTOCOL_DES,
-         '3desede',               PRIV_PROTOCOL_3DESEDE,
-         'cbc3desede',            PRIV_PROTOCOL_3DESEDE,
-         PRIV_PROTOCOL_3DESEDE,   PRIV_PROTOCOL_3DESEDE,
-         'aes128cfb',             PRIV_PROTOCOL_AESCFB128,
-         'aescfb128',             PRIV_PROTOCOL_AESCFB128,
-         PRIV_PROTOCOL_AESCFB128, PRIV_PROTOCOL_AESCFB128,
-         'aes192cfb',             PRIV_PROTOCOL_AESCFB192,
-         'aescfb192',             PRIV_PROTOCOL_AESCFB192,
-         PRIV_PROTOCOL_AESCFB192, PRIV_PROTOCOL_AESCFB192,
-         'aes256cfb',             PRIV_PROTOCOL_AESCFB256,
-         'aescfb256',             PRIV_PROTOCOL_AESCFB256,
-         PRIV_PROTOCOL_AESCFB256, PRIV_PROTOCOL_AESCFB256,
+         'des',                          PRIV_PROTOCOL_DES,
+         'cbcdes',                       PRIV_PROTOCOL_DES,
+         PRIV_PROTOCOL_DES,              PRIV_PROTOCOL_DES,
+         '3desede',                      PRIV_PROTOCOL_DRAFT_3DESEDE,
+         'cbc3desede',                   PRIV_PROTOCOL_DRAFT_3DESEDE,
+         PRIV_PROTOCOL_DRAFT_3DESEDE,    PRIV_PROTOCOL_DRAFT_3DESEDE,
+         'aes128cfb',                    PRIV_PROTOCOL_AESCFB128,
+         'aescfb128',                    PRIV_PROTOCOL_AESCFB128,
+         'cfbaes128',                    PRIV_PROTOCOL_AESCFB128,
+         'cfb128aes128',                 PRIV_PROTOCOL_AESCFB128,
+         PRIV_PROTOCOL_AESCFB128,        PRIV_PROTOCOL_AESCFB128,
+         PRIV_PROTOCOL_DRAFT_AESCFB128,  PRIV_PROTOCOL_AESCFB128,
+         'aes192cfb',                    PRIV_PROTOCOL_DRAFT_AESCFB192,
+         'aescfb192',                    PRIV_PROTOCOL_DRAFT_AESCFB192,
+         'cfbaes192',                    PRIV_PROTOCOL_DRAFT_AESCFB192,
+         'cfb128aes192',                 PRIV_PROTOCOL_DRAFT_AESCFB192,
+         PRIV_PROTOCOL_DRAFT_AESCFB192,  PRIV_PROTOCOL_DRAFT_AESCFB192,
+         'aes256cfb',                    PRIV_PROTOCOL_DRAFT_AESCFB256,
+         'aescfb256',                    PRIV_PROTOCOL_DRAFT_AESCFB256,
+         'cfbaes256',                    PRIV_PROTOCOL_DRAFT_AESCFB256,
+         'cfb128ae256',                  PRIV_PROTOCOL_DRAFT_AESCFB256,
+         PRIV_PROTOCOL_DRAFT_AESCFB256,  PRIV_PROTOCOL_DRAFT_AESCFB256,
       };
 
       if ($proto eq '') {
@@ -825,25 +833,32 @@ sub _priv_protocol
       my @match = grep(/^\Q$proto/i, keys(%{$protocols}));
 
       if (@match > 1) {
-         return $this->_error('Ambiguous privProtocol specified [%s]', $proto);
-      }
-
-      if (@match != 1) {
+         if (lc($proto) eq 'aes') {
+             $match[0] = 'aescfb128';
+         } else {
+             return $this->_error(
+                 'Ambiguous privProtocol specified [%s]', $proto
+             );
+         }
+      } elsif (@match != 1) {
          return $this->_error('Unknown privProtocol specified [%s]', $proto);
       }
 
       $this->{_priv_protocol} = $protocols->{$match[0]};
 
-      # Validate the support of the AES cipher algorithm.
+      # Validate the support of the AES cipher algorithm.  Attempt to 
+      # load the Crypt::Rijndael module.  If this module is not found, 
+      # do not provide support for the AES Cipher Algorithm.
 
-      if (defined($NO_AES_SUPPORT) &&
-          (($this->{_priv_protocol} eq PRIV_PROTOCOL_AESCFB128) ||
-           ($this->{_priv_protocol} eq PRIV_PROTOCOL_AESCFB192) ||
-           ($this->{_priv_protocol} eq PRIV_PROTOCOL_AESCFB256)))
+      if (($this->{_priv_protocol} eq PRIV_PROTOCOL_AESCFB128)       ||
+          ($this->{_priv_protocol} eq PRIV_PROTOCOL_DRAFT_AESCFB192) ||
+          ($this->{_priv_protocol} eq PRIV_PROTOCOL_DRAFT_AESCFB256))
       {
-         return $this->_error(
-            'No support for privProtocol [%s]' . $NO_AES_SUPPORT, $proto
-         ); 
+         if (defined(my $error = load_module('Crypt::Rijndael'))) {
+            return $this->_error(
+               'Support unavailable for privProtocol [%s] %s', $proto, $error 
+            );
+         }
       }
    }
 
@@ -1142,11 +1157,11 @@ sub _auth_data_init
 {
    my $encrypt = 
    {
-      PRIV_PROTOCOL_DES,       \&_priv_encrypt_des,          
-      PRIV_PROTOCOL_3DESEDE,   \&_priv_encrypt_3desede,
-      PRIV_PROTOCOL_AESCFB128, \&_priv_encrypt_aescfbxxx,
-      PRIV_PROTOCOL_AESCFB192, \&_priv_encrypt_aescfbxxx,
-      PRIV_PROTOCOL_AESCFB256, \&_priv_encrypt_aescfbxxx
+      PRIV_PROTOCOL_DES,              \&_priv_encrypt_des,          
+      PRIV_PROTOCOL_DRAFT_3DESEDE,    \&_priv_encrypt_3desede,
+      PRIV_PROTOCOL_AESCFB128,        \&_priv_encrypt_aescfbxxx,
+      PRIV_PROTOCOL_DRAFT_AESCFB192,  \&_priv_encrypt_aescfbxxx,
+      PRIV_PROTOCOL_DRAFT_AESCFB256,  \&_priv_encrypt_aescfbxxx
    };
 
    sub _encrypt_data
@@ -1175,11 +1190,11 @@ sub _auth_data_init
 {
    my $decrypt =
    {
-      PRIV_PROTOCOL_DES,       \&_priv_decrypt_des,
-      PRIV_PROTOCOL_3DESEDE,   \&_priv_decrypt_3desede,
-      PRIV_PROTOCOL_AESCFB128, \&_priv_decrypt_aescfbxxx,
-      PRIV_PROTOCOL_AESCFB192, \&_priv_decrypt_aescfbxxx,
-      PRIV_PROTOCOL_AESCFB256, \&_priv_decrypt_aescfbxxx
+      PRIV_PROTOCOL_DES,              \&_priv_decrypt_des,
+      PRIV_PROTOCOL_DRAFT_3DESEDE,    \&_priv_decrypt_3desede,
+      PRIV_PROTOCOL_AESCFB128,        \&_priv_decrypt_aescfbxxx,
+      PRIV_PROTOCOL_DRAFT_AESCFB192,  \&_priv_decrypt_aescfbxxx,
+      PRIV_PROTOCOL_DRAFT_AESCFB256,  \&_priv_decrypt_aescfbxxx
    };
 
    sub _decrypt_data
@@ -1209,12 +1224,15 @@ sub _auth_data_init
       }
       return $_[0]->_error($_[1]->error) unless ($_[1]->length);
 
-      # See if the decrypted data starts with a SEQUENCE
-      if (!defined($_[1]->process(SEQUENCE))) {
+      # See if the decrypted data starts with a SEQUENCE 
+      # and has a reasonable length.
+
+      my $msglen = $_[1]->process(SEQUENCE);
+      if ((!defined($msglen)) || ($msglen > $_[1]->length)) {
          return $_[0]->_error('Decryption error');
       }
       $_[1]->index(0); # Reset the index
-
+      
       DEBUG_INFO('privacy passed');
 
       TRUE;
@@ -1233,11 +1251,11 @@ sub _priv_data_init
 
    my $init =
    {
-      PRIV_PROTOCOL_DES,       \&_priv_data_init_des,
-      PRIV_PROTOCOL_3DESEDE,   \&_priv_data_init_3desede,
-      PRIV_PROTOCOL_AESCFB128, \&_priv_data_init_aescfbxxx,
-      PRIV_PROTOCOL_AESCFB192, \&_priv_data_init_aescfbxxx,
-      PRIV_PROTOCOL_AESCFB256, \&_priv_data_init_aescfbxxx
+      PRIV_PROTOCOL_DES,              \&_priv_data_init_des,
+      PRIV_PROTOCOL_DRAFT_3DESEDE,    \&_priv_data_init_3desede,
+      PRIV_PROTOCOL_AESCFB128,        \&_priv_data_init_aescfbxxx,
+      PRIV_PROTOCOL_DRAFT_AESCFB192,  \&_priv_data_init_aescfbxxx,
+      PRIV_PROTOCOL_DRAFT_AESCFB256,  \&_priv_data_init_aescfbxxx
    };
 
    if (!exists($init->{$this->{_priv_protocol}})) {
@@ -1391,7 +1409,7 @@ sub _priv_encrypt_3desede
    }
    $_[1] = pack('NN', $_[0]->{_engine_boots}, $_[0]->{_priv_data}->{salt});
 
-   # 3DES-EDE for USM Section 5.1.1.1.2 - "To achieve effective 
+   # Draft 3DES-EDE for USM Section 5.1.1.1.2 - "To achieve effective 
    # bit spreading, the complete 8-octet 'salt' value SHOULD be 
    # hashed using the usmUserAuthProtocol."
 
@@ -1433,7 +1451,9 @@ sub _priv_decrypt_3desede
    }
 
    if (length($_[2]) % 8) {
-      return $_[0]->_error('3DES-EDE cipher length not multiple of block size');
+      return $_[0]->_error(
+         'CBC-3DES-EDE cipher length not multiple of block size'
+      );
    }
 
    # Create the initial vector (IV)
@@ -1463,10 +1483,13 @@ sub _priv_data_init_aescfbxxx
       return $this->_error('Required privKey not defined');
    }
 
+   # Avoid a "strict subs" error if Crypt::Rijndael is not loaded.
+   no strict 'subs';
+
    # Create the AES (Rijndael) object with a 128, 192, or 256 bit key.
 
    $this->{_priv_data}->{aes} = 
-      Crypt::Rijndael->new($this->{_priv_key}, Crypt::Rijndael::MODE_CFB);
+      Crypt::Rijndael->new($this->{_priv_key}, Crypt::Rijndael::MODE_CFB());
 
    # Initialize the salt
    $this->{_priv_data}->{salt1} = int(rand(~0));
@@ -1625,19 +1648,19 @@ sub _priv_key_generate
 
    return $this->_error unless defined($this->{_priv_key});
 
-   if ($this->{_priv_protocol} eq PRIV_PROTOCOL_3DESEDE) {
+   if ($this->{_priv_protocol} eq PRIV_PROTOCOL_DRAFT_3DESEDE) {
 
-      # 3DES-EDE for USM Section 2.1 - "To acquire the necessary
-      # number of key bits, the password-to-key algorithm may be
-      # chained using its output as further input in order to
-      # generate an appropriate number of key bits."
+      # Draft 3DES-EDE for USM Section 2.1 - "To acquire the necessary 
+      # number of key bits, the password-to-key algorithm may be chained
+      # using its output as further input in order to generate an
+      # appropriate number of key bits."
 
       $this->{_priv_key} .= $this->_password_localize($this->{_priv_key}); 
 
-   } elsif (($this->{_priv_protocol} eq PRIV_PROTOCOL_AESCFB192) ||
-            ($this->{_priv_protocol} eq PRIV_PROTOCOL_AESCFB256))
+   } elsif (($this->{_priv_protocol} eq PRIV_PROTOCOL_DRAFT_AESCFB192) ||
+            ($this->{_priv_protocol} eq PRIV_PROTOCOL_DRAFT_AESCFB256))
    {
-      # AES in the USM Section 3.1.2.1 - "...if the size of the 
+      # Draft AES in the USM Section 3.1.2.1 - "...if the size of the 
       # localized key is not large enough to generate an encryption 
       # key... ...set Kul = Kul || Hnnn(Kul) where Hnnn is the hash 
       # function for the authentication protocol..."
@@ -1662,11 +1685,11 @@ sub _priv_key_generate
 
    my $key_len =
    {
-      PRIV_PROTOCOL_DES,        16,  # RFC 3414 Section 8.2.1
-      PRIV_PROTOCOL_3DESEDE,    32,  # 3DES-EDE for USM Section 5.2.1
-      PRIV_PROTOCOL_AESCFB128,  16,  # AES in the USM Section 3.2.1
-      PRIV_PROTOCOL_AESCFB192,  24,  # AES in the USM Section 3.2.1
-      PRIV_PROTOCOL_AESCFB256,  32   # AES in the USM Section 3.2.1
+      PRIV_PROTOCOL_DES,              16,  # RFC 3414 Section 8.2.1
+      PRIV_PROTOCOL_DRAFT_3DESEDE,    32,  # Draft 3DES for USM Section 5.2.1
+      PRIV_PROTOCOL_AESCFB128,        16,  # AES in the USM Section 3.2.1
+      PRIV_PROTOCOL_DRAFT_AESCFB192,  24,  # Draft AES in the USM Section 3.2.1
+      PRIV_PROTOCOL_DRAFT_AESCFB256,  32   # Draft AES in the USM Section 3.2.1
    };
 
    if (!exists($key_len->{$this->{_priv_protocol}})) {
@@ -1685,11 +1708,11 @@ sub _priv_key_validate
 
    my $key_len =
    {
-      PRIV_PROTOCOL_DES,        [ 16, 'DES'         ],  
-      PRIV_PROTOCOL_3DESEDE,    [ 32, '3DES-EDE'    ], 
-      PRIV_PROTOCOL_AESCFB128,  [ 16, 'AES-CFB-128' ],
-      PRIV_PROTOCOL_AESCFB192,  [ 24, 'AES-CFB-192' ],
-      PRIV_PROTOCOL_AESCFB256,  [ 32, 'AES-CFB-256' ]
+      PRIV_PROTOCOL_DES,              [ 16, 'CBC-DES'        ],  
+      PRIV_PROTOCOL_DRAFT_3DESEDE,    [ 32, 'CBC-3DES-EDE'   ], 
+      PRIV_PROTOCOL_AESCFB128,        [ 16, 'CFB128-AES-128' ],
+      PRIV_PROTOCOL_DRAFT_AESCFB192,  [ 24, 'CFB128-AES-192' ],
+      PRIV_PROTOCOL_DRAFT_AESCFB256,  [ 32, 'CFB128-AES-256' ]
    };
 
    if (!exists($key_len->{$this->{_priv_protocol}})) {
@@ -1707,26 +1730,26 @@ sub _priv_key_validate
       );
    }
   
-   if ($this->{_priv_protocol} eq PRIV_PROTOCOL_3DESEDE) {
+   if ($this->{_priv_protocol} eq PRIV_PROTOCOL_DRAFT_3DESEDE) {
 
-      # 3DES-EDE for USM Section 5.1.1.1.1 "The checks for difference 
+      # Draft 3DES-EDE for USM Section 5.1.1.1.1 "The checks for difference 
       # and weakness... ...should be performed when the key is assigned.
       # If any of the mandated tests fail, then the whole key MUST be 
       # discarded and an appropriate exception noted."
 
       if (substr($this->{_priv_key}, 0, 8) eq substr($this->{_priv_key}, 8, 8)) 
       {
-         return $this->_error('Invalid 3DES-EDE privKey [K1 equals K2]');
+         return $this->_error('Invalid CBC-3DES-EDE privKey [K1 equals K2]');
       }
 
       if (substr($this->{_priv_key}, 8, 8) eq substr($this->{_priv_key}, 16, 8))
       {
-         return $this->_error('Invalid 3DES-EDE privKey [K2 equals K3]');
+         return $this->_error('Invalid CBC-3DES-EDE privKey [K2 equals K3]');
       }
 
       if (substr($this->{_priv_key}, 0, 8) eq substr($this->{_priv_key}, 16, 8))
       {
-         return $this->_error('Invalid 3DES-EDE privKey [K1 equals K3]');
+         return $this->_error('Invalid CBC-3DES-EDE privKey [K1 equals K3]');
       }
 
    }
@@ -1767,18 +1790,49 @@ sub _password_localize
    $digest->add($d . $this->{_engine_id} . $d)->digest;
 }
 
+{
+   my %modules;
+
+   sub load_module
+   {
+      my ($module) = @_;
+
+      # We attempt to load the required module under the protection of an
+      # eval statement.  If there is a failure, typically it is due to a
+      # missing module required by the requested module and we attempt to
+      # simplify the error message by just listing that module.  We also
+      # need to track failures since require() only produces an error on
+      # the first attempt to load the module.
+
+      # NOTE: Contrary to our typical convention, a return value of "undef"
+      # actually means success and a defined value means error.
+
+      return $modules{$module} if (exists($modules{$module}));
+
+      if (!eval("require $module")) {
+         if ($@ =~ /locate (\S+\.pm)/) {
+            $modules{$module} = sprintf('(Required module %s not found)', $1);
+         } else {
+            $modules{$module} = sprintf('(%s)', $@);
+         }
+      } else {
+         $modules{$module} = undef;
+      }
+   }
+}
+
 sub DEBUG_INFO
 {
-   return unless $DEBUG;
+   return unless $Net::SNMP::Security::DEBUG;
 
    printf(
       sprintf('debug: [%d] %s(): ', (caller(0))[2], (caller(1))[3]) .
-      shift(@_) .
+      ((@_ > 1) ? shift(@_) : '%s') .
       "\n",
       @_
    );
 
-   $DEBUG;
+   $Net::SNMP::Security::DEBUG;
 }
 
 # ============================================================================
